"""Application Module defines application class used by other packages."""
import hashlib
import os
import re
import subprocess  # nosec
import sys
import tarfile
import yaml

from abc import ABC
from abc import abstractmethod

from app_gen_tool.common import extract
from app_gen_tool.common import get_chart_from_git
from app_gen_tool.common import to_camel_case
from app_gen_tool.common import transform_rel_path_into_abs_path
from app_gen_tool import constants


class Application(ABC):
    """Define Application Object."""

    def __init__(self, app_data: dict, output_folder: str):
        """Initialize class variables.

        Args:
            app_data (dict): Data from the input file.
            output_folder (str): Path to where the application will be generated.
        """
        # Initialize application config
        self._app: dict = {}
        self._app = app_data["appManifestFile-config"]
        self._temp_app_dir = os.path.join(constants.TEMP_USER_DIR,
                                          self.get_app_name())
        self.output_folder: str = output_folder

        self.APP_NAME = self._app["appName"]
        self.APP_NAME_WITH_UNDERSCORE = self._app["appName"].replace("-", "_")
        self.APP_NAME_CAMEL_CASE = (
            self._app["appName"].replace("-", " ").title().replace(" ", "")
        )

        # Initialize chartgroup
        self._chartgroup = app_data["appManifestFile-config"]["chartGroup"] \
            = app_data["appManifestFile-config"]["chart"][0]["chartGroup"]

        # Create empty list that will contain all the charts
        self._listcharts = {
            "chart_names": [],
            "namespace": self._app["namespace"]
        }

        # Initialize chart
        self._chart = app_data["appManifestFile-config"]["chart"]
        for i in range(len(self._chart)):
            self._chart[i]["namespace"] = self._app["namespace"]
            self._listcharts["chart_names"].append(self._chart[i]["name"])
            self._chart[i]['pathType'] = self._define_chart_pathtype(i)
            if self._chart[i]['pathType'] == 'git':
                self._process_git_helm_charts(i)
            else:
                self._chart[i]['path'] = transform_rel_path_into_abs_path(os.getcwd(),
                                                                          self._chart[i]['path'])
            if self._chart[i]['pathType'] == 'tarball':
                self._extract_helm_tars_to_temp_dir(chart_index=i)
            ret = self._get_values_file(self._chart[i]['path'])
            self._chart[i]['values'] = ret

        # Initialize setup data
        self.plugin_setup = app_data["setupFile-config"]

        # Initialize metadata
        self.metadata = app_data["metadataFile-config"]

    @abstractmethod
    def generate(self, package_only):
        """Abstract method for generating application."""

    def check_charts(self):
        """Subprocess that checks chart information."""
        charts = self._chart
        for chart in charts:
            manifest_data: dict = {}
            chart_file_data: dict = {}
            chart_yaml_file = os.path.join(chart["path"], "Chart.yaml")

            manifest_data["name"], manifest_data["version"] = (
                chart["name"],
                chart["version"],
            )

            try:
                chart_metadata_f = open(chart_yaml_file, "r",
                                        encoding="utf-8")
            except Exception as e:
                print(f"ERROR: {e}")
                sys.exit(1)
            chart_file_lines = chart_metadata_f.readlines()
            chart_file_lines = [
                cline
                for cline in chart_file_lines
                if len(cline) > 0 and cline[0] != "#"
            ]
            chart_metadata_f.close()
            for line in chart_file_lines:
                line = line.rstrip("\n")
                line_data = line.split()
                if not line_data:
                    continue
                if "name:" in line_data[0]:
                    chart_file_data["name"] = line_data[-1]
                elif "version:" in line_data[0]:
                    chart_file_data["version"] = line_data[-1]

            # To-do chart type different from dir
            for key in manifest_data:
                err_str = ""
                if key not in chart_file_data:
                    err_str = \
                        f'"{key}" is present in app-manifest.yaml but not in' \
                        f'{chart_yaml_file}'
                    raise KeyError(err_str)
                if manifest_data[key] != chart_file_data[key]:
                    err_str = \
                        f'"{key}" has different values ' \
                        f'in app-manifest.yaml and {chart_yaml_file}'
                    raise ValueError(err_str)

    def get_app_name(self):
        """Return name of app.

        Returns:
            str: Name of the application.
        """
        return self._app["appName"]

    @staticmethod
    def _package_helm_chart(chart, chart_dir) -> bool:
        """Sub-process of app generation. Generate application helm-charts tarball.

        Args:
            chart (dict): Data from the input file.
            chart_dir (str): Path to directory where the package chart will be placed.

        Returns:
            bool: True if Chart is succesfully linted and packaged. False otherwise.
        """
        # lint helm chart
        print(f"Linting Helm Chart: {chart['path']}")
        print(f"Chart Directory: {chart_dir}")

        cmd_lint = ["helm", "lint", chart["path"]]
        subproc = subprocess.run(
            cmd_lint,
            env=os.environ.copy(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if subproc.returncode == 0:
            print(str(subproc.stdout, encoding="utf-8"))
        else:
            print(str(subproc.stdout, encoding="utf-8"))
            print(str(subproc.stderr, encoding="utf-8"))
            return False

        # package helm chart
        cmd_package = ["helm", "package", chart["path"],
                       "--destination=" + chart_dir]
        subproc = subprocess.run(
            cmd_package,
            env=os.environ.copy(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if subproc.returncode == 0:
            output = str(subproc.stdout, encoding="utf-8")
            print(output)
            # capture tarball name
            for words in output.split("/"):
                if "tgz" in words:
                    chart["tarballName"] = words.rstrip("\n")
        else:
            print(subproc.stderr)
            return False

        return True

    @staticmethod
    def _get_values_file(path):
        """Retrieve the content of a values file and saves them in a dictionary.

        Args:
            path (str): Path to the Helm chart informed in the input file.

        Returns:
            dict: Data from the values.yaml found in the Helm chart path.
        """

        values_yaml_path = os.path.join(path, "values.yaml")
        try:
            with open(values_yaml_path, "r", encoding='utf-8') as f:
                values_yaml_file = yaml.safe_load(f)
        except FileNotFoundError:
            print(f'File {values_yaml_path} not found')
            sys.exit(1)
        except yaml.YAMLError:
            print(f'Error while trying to load {values_yaml_path}, '
                  'check if it is a valid YAML file.')
            sys.exit(1)
        return values_yaml_file

    # pyyaml does not support writing yaml block with initial indent
    # add initial indent for yaml block substitution
    @staticmethod
    def _write_yaml_to_manifest(key, src, init_indent):
        """Return the values of a given key in a yaml file.

        Args:
            key (str): Key that the values will be returned.
            src (list): List of values that corresponds to the given key.
            init_indent (int): Indentation to be added to the return value.

        Returns:
            str: Values with indentation and break lines.
        """
        target = {key: src}
        # add heading key
        lines = yaml.safe_dump(target, sort_keys=False).split("\n")
        # remove ending space and first line
        lines.pop()
        lines.pop(0)
        indents = " " * init_indent
        for i in range(len(lines)):
            lines[i] = indents + lines[i]
        # restore ending '\n'
        return "\n".join(lines) + "\n"

    def _write_values_and_blocks(self, file, line: str, chart):
        """Substitute values in a template file.

        Args:
            file (str): File that the new line will be added.
            line (str): Line that will have values replaced.
            chart (dict): Data from the input file.
        """
        # substitute template values to chart values
        out_line, substituted = self._substitute_values(line, chart)
        if not substituted:
            # substitute template blocks to chart blocks
            out_line = self._substitute_blocks(line, chart)
        file.write(out_line)

    def _substitute_values(self, in_line, dicts):
        """Substitute value in a string.

        Args:
            in_line (str): Line that will have values replaced.
            dicts (dict): Data from the input file.

        Returns:
            tuple (str,bool): Line with/without value substituted and True if substitution ocurred.
        """
        out_line = in_line
        pattern = re.compile(r'\$.+?\$')
        results = pattern.findall(out_line)
        if results:
            for result in results:
                result_word = result.strip("$").split("%")
                value_key = result_word[0]
                value_default = ""
                if len(result_word) > 1:
                    value_default = result_word[1]
                # underscore case to camel case
                value = to_camel_case(value_key)
                if value in dicts:
                    out_line = out_line.replace(result, str(dicts[value]))
                elif value_default:
                    out_line = out_line.replace(result, value_default)

        if out_line == in_line:
            return out_line, False
        else:
            return out_line, True

    def _substitute_blocks(self, in_line, dicts):
        """Substitute a value with a block of data in a string.

        Args:
            in_line (str): Line that will have values replaced.
            dicts (dict): Data from the input file.

        Returns:
            str: Line with or without a value substituted.
        """
        out_line = in_line
        result = re.search(r'@\S+\|\d+@', out_line)
        if result:
            block_key = result.group().strip("@").split("|")
            key = block_key[0].lower()
            indent = int(block_key[1])
            if key in dicts:
                out_line = self._write_yaml_to_manifest(key, dicts[key],
                                                        indent)
            else:
                out_line = ""
        return out_line

    # TODO: sub-chart dependency check
    def _gen_helm_chart_tarball(self, chart, chart_dir):
        """Sub-process of app generation. Lint and package helm chart.

        Args:
            chart (dict): Data from the input file.
            chart_dir (str): Path to where the packaged Helm chart will be placed.

        Returns:
            bool: True if the Helm chart was succesfully packaged. False otherwise.
        """
        ret = False
        chart_path = ""
        print(f'Processing chart {chart["name"]}...')
        chart_path = chart["path"]
        # update chart path
        # remove ending '/', if it exists...
        chart["path"] = chart_path.rstrip("/")
        # lint and package
        ret = self._package_helm_chart(chart, chart_dir)

        return ret

    # Sub-process of app generation
    # generate application metadata
    def _gen_metadata(self, output_dir):
        """Sub-process of app generation. Generate application metadata.

        Args:
            output_dir (str): Path to where the metadata file will be placed.

        Returns:
            bool: True if the metadata was succesfully created. False otherwise.
        """
        # gets the keys and values defined in the input yaml and writes the metadata.yaml app file.

        app_name, app_version = self._app["appName"], self._app["appVersion"]
        metadata_file = os.path.join(output_dir, "metadata.yaml")

        try:
            with open(metadata_file, "w", encoding="utf-8") as f:
                f.write(f"app_name: {app_name}\napp_version: {app_version}")
        except Exception:
            return False

        return True

    # Sub-process of app generation
    # generate application sha256 file
    @staticmethod
    def _gen_sha256(in_file):
        """Sub-process of app generation. Generate hash function of a file.

        Args:
            in_file (str): Path to the file that the hash function will be generated to.

        Returns:
            str: Hash function value.
        """
        with open(in_file, "rb") as f:
            out_sha256 = hashlib.sha256(f.read()).hexdigest()
        return out_sha256

    # Sub-process of app generation
    # generate application checksum file and tarball
    def _gen_checksum_and_app_tarball(self, output):
        """Sub-process of app generation. Generate checksum file and application package.

        Args:
            output (str): Path to where that packaged application will be placed.

        Returns:
            str: Name of the generated tarball.
        """
        store_cwd = os.getcwd()
        app_files = []
        os.chdir(output)

        # gen checksum
        # check checksum file existance
        checksum_file = "checksum.sha256"
        if os.path.exists(checksum_file):
            os.remove(checksum_file)

        try:
            for parent, _, filenames in os.walk("./"):
                for filename in filenames:
                    if filename[-3:] != ".py" and filename[-4:] != ".cfg":
                        app_files.append(os.path.join(parent, filename))
        except Exception as e:
            print(f"Error: {e}")

        try:
            with open(checksum_file, "a", encoding="utf-8") as f:
                for target_file in sorted(app_files):
                    f.write(self._gen_sha256(
                        target_file) + " *" + target_file + "\n")
        except Exception as e:
            print(f"Error: {e}")

        app_files.append("./" + checksum_file)

        # gen application tarball
        tarname = f"{self._app['appName']}-{self._app['appVersion']}.tgz"

        with tarfile.open(tarname, "w:gz") as t:
            for target_file in app_files:
                t.add(target_file)

            t.close()

        os.chdir(store_cwd)
        return tarname

    def _extract_helm_tars_to_temp_dir(self, chart_index):
        """Auxiliary function to extract helm charts passed as tar packages"""

        try:
            chart = self._chart[chart_index]
            tarball_dir = chart['path']
            tmp_rel_path = self._temp_app_dir
            extract_path = os.path.join(tmp_rel_path, self.APP_NAME)
            if not os.path.exists(extract_path):
                os.makedirs(extract_path)
            extract(tarball_dir, extract_path)
            self._chart[chart_index]['tarball_path'] = \
                chart['path']
            self._chart[chart_index]['path'] = os.path.join(extract_path,
                                                            chart['name'])
        except Exception as e:
            print(f"ERROR: {e}")
            sys.exit(1)

    def _define_chart_pathtype(self, i: int):
        """Gets pathtype for each chart"""
        chart = self._chart[i]
        path = chart['path']
        path_type = ''
        if path.endswith('.tar') or path.endswith('.tar.gz') \
                or path.endswith('.tgz'):
            path_type = 'tarball'
        elif path.endswith('.git'):
            path_type = 'git'
        else:
            path_type = 'dir'
        return path_type

    def _process_git_helm_charts(self, chart_index: int):
        self._chart[chart_index]['_gitname'] = re.search(r'[^/]+(?=\.git$)',
                                                         self._chart[
                                                             chart_index][
                                                             'path']).group()
        chart = self._chart[chart_index]
        if not os.path.exists(self._temp_app_dir):
            os.makedirs(self._temp_app_dir)
        # if the git folder exists, check git name and use that folder
        # otherwise git clone from upstream
        new_path = os.path.join(self._temp_app_dir,
                                chart['_gitname'])

        _, saved_pwd = get_chart_from_git(
            new_path, self._temp_app_dir,
            chart['path'],
            chart['_gitname']
        )
        self._chart[chart_index]['_gitUrl'] = chart['path']
        self._chart[chart_index]['path'] = os.path.join(saved_pwd, new_path)
        os.chdir(saved_pwd)
        return True

    def print_app_data(self):
        """Debug Print App Data."""
        print(self._app)
        print(self._chartgroup)
        print(self._chart)
