"""Module for common methods accross entire codebase."""
import os
import subprocess
import sys
import tarfile


def to_camel_case(s: str) -> str:
    """Convert String to camel case.

    Args:
        s (str): String that will be converted to camel case.

    Returns:
        str: String converted to camel case.
    """
    return s[0].lower() + s.title().replace('_', '')[1:] if s else s


def uppercase_name(name: str) -> str:
    """Transform all letters of a name in uppercase.

    Args:
        name (str): String that will be converted to uppercase.

    Returns:
        str: String converted to uppercase.
    """
    return name.upper().replace(" ", "_").replace("-", "_")


def extract(tar_path='.', extract_path='.'):
    """Extracts content from a .tar package stored locally or on the web"""

    try:
        if not os.path.exists(tar_path):
            new_path = os.getcwd()
            tar_path = transform_rel_path_into_abs_path(new_path, tar_path)
        tar = tarfile.open(tar_path, 'r')
        for item in tar:
            tar.extract(item, extract_path)
            if item.name.find(".tgz") != -1 or item.name.find(".tar") != -1:
                extract(item.name, "./" + item.name[:item.name.rfind('/')])
    except Exception as e:
        print(f"Exception error {e} occurred during tar extraction!")
        sys.exit(1)


def get_chart_from_git(new_path: str, temp_dir: str, git_url: str, git_name: str):
    """Auxiliary function to clone helm-charts from a git url

    Returns a pair of type boolean, string.
    The boolean indicates whether the git clone was successful or not.
    The string indicates the saved pwd
    """
    if not os.path.exists(new_path):
        saved_pwd = os.getcwd()
        os.chdir(temp_dir)
        cmd = ["git", "clone", git_url]

        subproc = subprocess.run(
            cmd,
            env=os.environ.copy(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if subproc.returncode != 0:
            output = str(subproc.stderr, encoding="utf-8")
            print(output)
            print(f'Error: git clone {git_name} failed')
            os.chdir(saved_pwd)
            return False, saved_pwd
        os.chdir(saved_pwd)
    else:
        # git pull to ensure folder up-to-date
        saved_pwd = os.getcwd()
        os.chdir(new_path)
        cmd = ["git", "pull"]

        subproc = subprocess.run(
            cmd,
            env=os.environ.copy(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if subproc.returncode != 0:
            output = str(subproc.stderr, encoding="utf-8")
            print(output)
            print(f'Error: git pull for {git_name} failed')
            os.chdir(saved_pwd)
            return False, saved_pwd
    return True, saved_pwd


def transform_rel_path_into_abs_path(abs_path: str, rel_path: str) -> str:
    """It transforms the rel_path into an absolute path"""
    if rel_path[:2] == './':
        rel_path = rel_path.replace('./', 'app-gen-tool/')
    longest_common_path = ""
    for i in range(len(rel_path)):
        common_path = rel_path[:i + 1]
        if common_path in abs_path:
            longest_common_path = common_path
    replace_index = abs_path.find(longest_common_path)
    return abs_path[:replace_index] + rel_path
