"""Generator Main Module."""
import argparse
import os
import shutil
import sys
import yaml

from app_gen_tool.fluxcd import FluxCD


def parse_yaml(yaml_in):
    """Parse generator input yaml file.

    Args:
        yaml_in (str): Path to the input file.

    Returns:
        dict: Contents from the input file.
    """
    yaml_data = ''
    try:
        with open(yaml_in, 'r', encoding='utf-8') as f:
            yaml_data = yaml.safe_load(f)
    except FileNotFoundError:
        print('Error: {yaml_in} not found')
    except Exception:
        print('Error: Invalid yaml file')
    return yaml_data


def check_manifest(manifest_data):  # pylint: disable=too-many-return-statements
    """Check generator input yaml file for correct inputs.

    Args:
        manifest_data (dict): contens from the input file.

    Returns:
        bool: False if any required field is missing from input file. True otherwise.
    """
    # TODO: check more mandatory key/values in manifest yaml

    # check app values
    if 'appName' not in manifest_data['appManifestFile-config']:
        print('Error: "appName" is missing.')
        return False

    if 'namespace' not in manifest_data['appManifestFile-config']:
        print('Error: "namespace" is missing.')
        return False

    if 'appVersion' not in manifest_data['appManifestFile-config']:
        print('Error: "appVersion" is missing.')
        return False

    # check chartGroup values
    try:
        if 'chartGroup' not in \
                manifest_data['appManifestFile-config']['chart'][0]:
            print('Error: "chartGroup" is missing.')
            return False
    except IndexError:
        print("Please make sure that there is at least one chart"
              "correctly specified in the app-gen tool")
        return False

    # check chart values
    if 'chart' not in manifest_data['appManifestFile-config']:
        print('Error: "chart" is missing.')
        return False

    for chart in manifest_data['appManifestFile-config']['chart']:
        # check chart name
        if 'name' not in chart:
            print('Error: Chart attribute "name" is missing.')
            return False

        # check chart version
        if 'version' not in chart:
            print('Error: Chart attribute "version" is missing.')
            return False

        # check chart path, supporting: dir, git, tarball
        if 'path' not in chart:
            print(
                f'Error: Chart attribute \'path\' is missing in chart {chart["name"]}.'
            )
            return False
    return True


def check_input_file(app_data) -> bool:
    """Check if input file passed by user is valid.

    Args:
        app_data (dict): Contents from the input file.

    Returns:
        bool: True if input file is a valid YAML file. False otherwise.
    """
    if not app_data:
        print('Parse yaml error')
        return False
    if not check_manifest(app_data):
        print('Application manifest is not valid')
        return False
    return True


def check_app_directory(app_out, overwrite, package_only) -> bool:
    """Check if the user gave enough information to modify or create app folder directory.

    Args:
        app_out (str): Path to where the application files will be placed.
        overwrite (bool): Instructs the generator to delete, if already exists, the output dir.
        package_only (bool): Instructs the generator to only execute packaging related code.

    Returns:
        bool: False if the arguments for creating or modifying where passed. True otherwise.
    """
    if os.path.exists(app_out) and not overwrite and not package_only:
        print(
            f'Output folder {app_out} exists, please remove it, use '
            f'--overwrite, or use --package-only.'
        )
        return False
    if not os.path.exists(app_out) and package_only:
        print(f'Output folder {app_out} does not exist, cannot package it.')
        return False
    return True


def extract_app_name(app_data):
    """Extract the app name from the input yaml file.

    Args:
        app_data (dict): Contens from the input file.

    Returns:
        str: Name of the application.
    """
    return app_data['appManifestFile-config']['appName']


def create_app_directories(app_out, overwrite):
    """Create or/and remove the app generated directories.

    Args:
        app_out (str): Path to where the application files will be placed.
        overwrite (bool): Instructs the generator to delete, if already exists, the output dir.
    """
    if os.path.exists(app_out) and overwrite:
        shutil.rmtree(app_out)
    if not os.path.exists(app_out):
        os.makedirs(app_out)


def generate_app(app_data, output_folder, package_only):
    """Generate the application.

    Args:
        app_data (dict): Contents from the input file.
        output_folder (str): Path to where the application files will be placed.
        package_only (bool): Instructs the generator to only execute packaging related code.
    """
    app = FluxCD(app_data, output_folder)
    app.generate(package_only)


def handle_args():
    """Handle arguments supplied via command line."""
    parser = argparse.ArgumentParser(
        prog="stx-app-generator",
        description="StarlingX User Application Generator"
    )

    parser.add_argument(
        "-i", "--input", dest="input",
        help="""
        Path to the yaml file to be used as source of information for the application
        """,
        type=str,
        metavar="YAML_FILE"
    )

    parser.add_argument(
        "-o", "--output", dest="output",
        help="""
        Path to the output folder that all files created by the generator will be placed
        """,
        type=str,
        metavar="PATH"
    )

    parser.add_argument(
        "--overwrite",
        help="""
        Instructs the generator to overwrite the output folder in case already
        exists
        """,
        action='store_true'
    )

    parser.add_argument(
        "--package-only",
        help="""
        Instructs the generator to only execute packaging related actions with
        the already existing files. Not to be used in conjuntion with
        --overwrite
        """,
        action='store_true'
    )

    args = parser.parse_args()
    return args


def main():
    """Main Method with argument parsing."""
    ARGS = handle_args()

    if ARGS.overwrite and ARGS.package_only:
        print(
            'Error: Selecting both overwrite and package-only is not allowed'
            '. Please consult our README if further clarification is needed'
        )
        sys.exit(1)

    if not os.path.isfile(os.path.abspath(ARGS.input)):
        print('Error: input file not found')
        sys.exit(1)

    if ARGS.input:
        app_data = parse_yaml(ARGS.input)
        if not check_input_file(app_data):
            sys.exit(1)

        if ARGS.output:
            output_folder = os.path.join(
                os.path.abspath(ARGS.output), extract_app_name(app_data)
            )
        else:
            output_folder = os.path.join(
                ".", extract_app_name(app_data)
            )

        if not check_app_directory(
                output_folder, ARGS.overwrite, ARGS.package_only):
            sys.exit(1)

        create_app_directories(output_folder, ARGS.overwrite)

        generate_app(
            app_data, output_folder, ARGS.package_only)
